@@ -70,6 +70,7 @@ pub enum Expression {
     Scalar(Span, Number),
     Identifier(Span, String),
     UnitIdentifier(Span, Prefix, String, String),
+    TypedHole(Span),
     UnaryOperator {
         op: UnaryOperator,
         expr: Box<Expression>,
@@ -92,6 +93,7 @@ pub enum Expression {
         fields: Vec<(Span, String, Expression)>,
     },
     AccessField(Span, Span, Box<Expression>, String),
+    List(Span, Vec<Expression>),
 }
 
 impl Expression {
@@ -125,6 +127,8 @@ impl Expression {
             Expression::String(span, _) => *span,
             Expression::InstantiateStruct { full_span, .. } => *full_span,
             Expression::AccessField(full_span, _ident_span, _, _) => *full_span,
+            Expression::List(span, _) => *span,
+            Expression::TypedHole(span) => *span,
         }
     }
 }
@@ -221,6 +225,16 @@ macro_rules! struct_ {
     }};
 }
 
+#[cfg(test)]
+macro_rules! list {
+    ( $( $val:expr ),* ) => {
+        crate::ast::Expression::List(
+             Span::dummy(),
+            vec![$($val,)*],
+        )
+    };
+}
+
 #[cfg(test)]
 pub(crate) use binop;
 #[cfg(test)]
@@ -232,6 +246,8 @@ pub(crate) use factorial;
 #[cfg(test)]
 pub(crate) use identifier;
 #[cfg(test)]
+pub(crate) use list;
+#[cfg(test)]
 pub(crate) use logical_neg;
 #[cfg(test)]
 pub(crate) use negate;
@@ -242,23 +258,23 @@ pub(crate) use struct_;
 
 #[derive(Debug, Clone, PartialEq)]
 pub enum TypeAnnotation {
-    Never(Span),
     TypeExpression(TypeExpression),
     Bool(Span),
     String(Span),
     DateTime(Span),
     Fn(Span, Vec<TypeAnnotation>, Box<TypeAnnotation>),
+    List(Span, Box<TypeAnnotation>),
 }
 
 impl TypeAnnotation {
     pub fn full_span(&self) -> Span {
         match self {
-            TypeAnnotation::Never(span) => *span,
             TypeAnnotation::TypeExpression(d) => d.full_span(),
             TypeAnnotation::Bool(span) => *span,
             TypeAnnotation::String(span) => *span,
             TypeAnnotation::DateTime(span) => *span,
             TypeAnnotation::Fn(span, _, _) => *span,
+            TypeAnnotation::List(span, _) => *span,
         }
     }
 }
@@ -266,7 +282,6 @@ impl TypeAnnotation {
 impl PrettyPrint for TypeAnnotation {
     fn pretty_print(&self) -> Markup {
         match self {
-            TypeAnnotation::Never(_) => m::type_identifier("!"),
             TypeAnnotation::TypeExpression(d) => d.pretty_print(),
             TypeAnnotation::Bool(_) => m::type_identifier("Bool"),
             TypeAnnotation::String(_) => m::type_identifier("String"),
@@ -286,6 +301,12 @@ impl PrettyPrint for TypeAnnotation {
                     + return_type.pretty_print()
                     + m::operator("]")
             }
+            TypeAnnotation::List(_, element_type) => {
+                m::type_identifier("List")
+                    + m::operator("<")
+                    + element_type.pretty_print()
+                    + m::operator(">")
+            }
         }
     }
 }
@@ -367,25 +388,34 @@ pub enum ProcedureKind {
     Type,
 }
 
+#[derive(Debug, Clone, PartialEq)]
+pub enum TypeParameterBound {
+    Dim,
+}
+
+#[derive(Debug, Clone, PartialEq)]
+pub struct DefineVariable {
+    pub identifier_span: Span,
+    pub identifier: String,
+    pub expr: Expression,
+    pub type_annotation: Option<TypeAnnotation>,
+    pub decorators: Vec<Decorator>,
+}
+
 #[derive(Debug, Clone, PartialEq)]
 pub enum Statement {
     Expression(Expression),
-    DefineVariable {
-        identifier_span: Span,
-        identifier: String,
-        expr: Expression,
-        type_annotation: Option<TypeAnnotation>,
-        decorators: Vec<Decorator>,
-    },
+    DefineVariable(DefineVariable),
     DefineFunction {
         function_name_span: Span,
         function_name: String,
-        type_parameters: Vec<(Span, String)>,
-        /// Parameters, optionally with type annotations. The boolean argument specifies whether or not the parameter is variadic
-        parameters: Vec<(Span, String, Option<TypeAnnotation>, bool)>,
+        type_parameters: Vec<(Span, String, Option<TypeParameterBound>)>,
+        /// Parameters, optionally with type annotations.
+        parameters: Vec<(Span, String, Option<TypeAnnotation>)>,
         /// Function body. If it is absent, the function is implemented via FFI
         body: Option<Expression>,
-        return_type_annotation_span: Option<Span>,
+        /// Local variables
+        local_variables: Vec<DefineVariable>,
         /// Optional annotated return type
         return_type_annotation: Option<TypeAnnotation>,
         decorators: Vec<Decorator>,
@@ -397,7 +427,7 @@ pub enum Statement {
         identifier: String,
         expr: Expression,
         type_annotation_span: Option<Span>,
-        type_annotation: Option<TypeExpression>,
+        type_annotation: Option<TypeAnnotation>,
         decorators: Vec<Decorator>,
     },
     ProcedureCall(Span, ProcedureKind, Vec<Expression>),
@@ -418,7 +448,6 @@ pub trait ReplaceSpans {
 impl ReplaceSpans for TypeAnnotation {
     fn replace_spans(&self) -> Self {
         match self {
-            TypeAnnotation::Never(_) => TypeAnnotation::Never(Span::dummy()),
             TypeAnnotation::TypeExpression(d) => TypeAnnotation::TypeExpression(d.replace_spans()),
             TypeAnnotation::Bool(_) => TypeAnnotation::Bool(Span::dummy()),
             TypeAnnotation::String(_) => TypeAnnotation::String(Span::dummy()),
@@ -426,6 +455,9 @@ impl ReplaceSpans for TypeAnnotation {
             TypeAnnotation::Fn(_, pt, rt) => {
                 TypeAnnotation::Fn(Span::dummy(), pt.clone(), rt.clone())
             }
+            TypeAnnotation::List(_, et) => {
+                TypeAnnotation::List(Span::dummy(), Box::new(et.replace_spans()))
+            }
         }
     }
 }
@@ -537,6 +569,24 @@ impl ReplaceSpans for Expression {
                 Box::new(expr.replace_spans()),
                 attr.clone(),
             ),
+            Expression::List(_, elements) => Expression::List(
+                Span::dummy(),
+                elements.iter().map(|e| e.replace_spans()).collect(),
+            ),
+            Expression::TypedHole(_) => Expression::TypedHole(Span::dummy()),
+        }
+    }
+}
+
+#[cfg(test)]
+impl ReplaceSpans for DefineVariable {
+    fn replace_spans(&self) -> Self {
+        Self {
+            identifier_span: Span::dummy(),
+            identifier: self.identifier.clone(),
+            expr: self.expr.replace_spans(),
+            type_annotation: self.type_annotation.as_ref().map(|t| t.replace_spans()),
+            decorators: self.decorators.clone(),
         }
     }
 }
@@ -546,26 +596,16 @@ impl ReplaceSpans for Statement {
     fn replace_spans(&self) -> Self {
         match self {
             Statement::Expression(expr) => Statement::Expression(expr.replace_spans()),
-            Statement::DefineVariable {
-                identifier_span: _,
-                identifier,
-                expr,
-                type_annotation,
-                decorators,
-            } => Statement::DefineVariable {
-                identifier_span: Span::dummy(),
-                identifier: identifier.clone(),
-                expr: expr.replace_spans(),
-                type_annotation: type_annotation.as_ref().map(|t| t.replace_spans()),
-                decorators: decorators.clone(),
-            },
+            Statement::DefineVariable(variable) => {
+                Statement::DefineVariable(variable.replace_spans())
+            }
             Statement::DefineFunction {
                 function_name_span: _,
                 function_name,
                 type_parameters,
                 parameters,
                 body,
-                return_type_annotation_span: return_type_span,
+                local_variables,
                 return_type_annotation,
                 decorators,
             } => Statement::DefineFunction {
@@ -573,21 +613,23 @@ impl ReplaceSpans for Statement {
                 function_name: function_name.clone(),
                 type_parameters: type_parameters
                     .iter()
-                    .map(|(_, name)| (Span::dummy(), name.clone()))
+                    .map(|(_, name, bound)| (Span::dummy(), name.clone(), bound.clone()))
                     .collect(),
                 parameters: parameters
                     .iter()
-                    .map(|(_, name, type_, is_variadic)| {
+                    .map(|(_, name, type_)| {
                         (
                             Span::dummy(),
                             name.clone(),
                             type_.as_ref().map(|t| t.replace_spans()),
-                            *is_variadic,
                         )
                     })
                     .collect(),
                 body: body.clone().map(|b| b.replace_spans()),
-                return_type_annotation_span: return_type_span.map(|_| Span::dummy()),
+                local_variables: local_variables
+                    .iter()
+                    .map(DefineVariable::replace_spans)
+                    .collect(),
                 return_type_annotation: return_type_annotation.as_ref().map(|t| t.replace_spans()),
                 decorators: decorators.clone(),
             },
@@ -633,7 +675,7 @@ impl ReplaceSpans for Statement {
                 struct_name_span: Span::dummy(),
                 struct_name: struct_name.clone(),
                 fields: fields
-                    .into_iter()
+                    .iter()
                     .map(|(_span, name, type_)| {
                         (Span::dummy(), name.clone(), type_.replace_spans())
                     })
