@@ -4,6 +4,7 @@ use itertools::Itertools;
 
 use crate::ast::ProcedureKind;
 use crate::decorator::Decorator;
+use crate::dimension::DimensionRegistry;
 use crate::interpreter::{
     Interpreter, InterpreterResult, InterpreterSettings, Result, RuntimeError,
 };
@@ -11,12 +12,14 @@ use crate::name_resolution::LAST_RESULT_IDENTIFIERS;
 use crate::prefix::Prefix;
 use crate::prefix_parser::AcceptsPrefix;
 use crate::pretty_print::PrettyPrint;
-use crate::typed_ast::{BinaryOperator, Expression, Statement, StringPart, UnaryOperator};
+use crate::typed_ast::{
+    BinaryOperator, DefineVariable, Expression, Statement, StringPart, UnaryOperator,
+};
 use crate::unit::{CanonicalName, Unit};
 use crate::unit_registry::{UnitMetadata, UnitRegistry};
 use crate::value::FunctionReference;
 use crate::vm::{Constant, ExecutionContext, Op, Vm};
-use crate::{decorator, ffi};
+use crate::{decorator, ffi, Type};
 
 #[derive(Debug, Clone, Default)]
 pub struct LocalMetadata {
@@ -47,7 +50,7 @@ pub struct BytecodeInterpreter {
 impl BytecodeInterpreter {
     fn compile_expression(&mut self, expr: &Expression) -> Result<()> {
         match expr {
-            Expression::Scalar(_span, n) => {
+            Expression::Scalar(_span, n, _type) => {
                 let index = self.vm.add_constant(Constant::Scalar(n.to_f64()));
                 self.vm.add_op1(Op::LoadConstant, index);
             }
@@ -186,9 +189,15 @@ impl BytecodeInterpreter {
                 self.vm
                     .add_op2(Op::BuildStructInstance, struct_info_idx, exprs.len() as u16);
             }
-            Expression::AccessField(_span, _full_span, expr, attr, struct_info, _result_type) => {
+            Expression::AccessField(_span, _full_span, expr, attr, struct_type, _result_type) => {
                 self.compile_expression_with_simplify(expr)?;
 
+                let Type::Struct(ref struct_info) = struct_type.to_concrete_type() else {
+                    unreachable!(
+                        "Field access of non-struct type should be prevented by the type checker"
+                    );
+                };
+
                 let idx = struct_info.fields.get_index_of(attr).unwrap();
 
                 self.vm.add_op1(Op::AccessStructField, idx as u16);
@@ -252,6 +261,16 @@ impl BytecodeInterpreter {
                 self.vm
                     .patch_u16_value_at(else_jump_offset, end_offset - (else_jump_offset + 2));
             }
+            Expression::List(_, elements, _) => {
+                for element in elements {
+                    self.compile_expression_with_simplify(element)?;
+                }
+
+                self.vm.add_op1(Op::BuildList, elements.len() as u16);
+            }
+            Expression::TypedHole(_, _) => {
+                unreachable!("Typed holes cause type inference errors")
+            }
         };
 
         Ok(())
@@ -272,45 +291,58 @@ impl BytecodeInterpreter {
             | Expression::String(..)
             | Expression::Condition(..)
             | Expression::InstantiateStruct(..)
-            | Expression::AccessField(..) => {}
+            | Expression::AccessField(..)
+            | Expression::List(..) => {}
             Expression::BinaryOperator(..) | Expression::BinaryOperatorForDate(..) => {
                 self.vm.add_op(Op::FullSimplify);
             }
+            Expression::TypedHole(_, _) => unreachable!("Typed holes cause type inference errors"),
         }
 
         Ok(())
     }
 
-    fn compile_statement(&mut self, stmt: &Statement) -> Result<()> {
+    fn compile_define_variable(&mut self, define_variable: &DefineVariable) -> Result<()> {
+        let DefineVariable(identifier, decorators, expr, _annotation, _type, _readable_type) =
+            define_variable;
+        let current_depth = self.current_depth();
+
+        // For variables, we ignore the prefix info and only use the names
+        let aliases = crate::decorator::name_and_aliases(identifier, decorators)
+            .map(|(name, _)| name)
+            .cloned()
+            .collect::<Vec<_>>();
+        let metadata = LocalMetadata {
+            name: crate::decorator::name(decorators),
+            url: crate::decorator::url(decorators),
+            description: crate::decorator::description(decorators),
+            aliases: aliases.clone(),
+        };
+
+        for alias_name in aliases {
+            self.compile_expression_with_simplify(expr)?;
+
+            self.locals[current_depth].push(Local {
+                identifier: alias_name.clone(),
+                depth: current_depth,
+                metadata: metadata.clone(),
+            });
+        }
+        Ok(())
+    }
+
+    fn compile_statement(
+        &mut self,
+        stmt: &Statement,
+        dimension_registry: &DimensionRegistry,
+    ) -> Result<()> {
         match stmt {
             Statement::Expression(expr) => {
                 self.compile_expression_with_simplify(expr)?;
                 self.vm.add_op(Op::Return);
             }
-            Statement::DefineVariable(identifier, decorators, expr, _type_annotation, _type) => {
-                let current_depth = self.current_depth();
-
-                // For variables, we ignore the prefix info and only use the names
-                let aliases = crate::decorator::name_and_aliases(identifier, decorators)
-                    .map(|(name, _)| name)
-                    .cloned()
-                    .collect::<Vec<_>>();
-                let metadata = LocalMetadata {
-                    name: crate::decorator::name(decorators),
-                    url: crate::decorator::url(decorators),
-                    description: crate::decorator::description(decorators),
-                    aliases: aliases.clone(),
-                };
-
-                for alias_name in aliases {
-                    self.compile_expression_with_simplify(expr)?;
-
-                    self.locals[current_depth].push(Local {
-                        identifier: alias_name.clone(),
-                        depth: 0,
-                        metadata: metadata.clone(),
-                    });
-                }
+            Statement::DefineVariable(define_variable) => {
+                self.compile_define_variable(define_variable)?
             }
             Statement::DefineFunction(
                 name,
@@ -318,8 +350,10 @@ impl BytecodeInterpreter {
                 _type_parameters,
                 parameters,
                 Some(expr),
-                _return_type_annotation,
+                local_variables,
                 _return_type,
+                _return_type_annotation,
+                _readable_return_type,
             ) => {
                 self.vm.begin_function(name);
 
@@ -333,6 +367,9 @@ impl BytecodeInterpreter {
                         metadata: LocalMetadata::default(),
                     });
                 }
+                for local_variables in local_variables {
+                    self.compile_define_variable(local_variables)?;
+                }
 
                 self.compile_expression_with_simplify(expr)?;
                 self.vm.add_op(Op::Return);
@@ -349,22 +386,16 @@ impl BytecodeInterpreter {
                 _type_parameters,
                 parameters,
                 None,
-                _return_type_annotation,
+                _local_variables,
                 _return_type,
+                _return_type_annotation,
+                _readable_return_type,
             ) => {
                 // Declaring a foreign function does not generate any bytecode. But we register
                 // its name and arity here to be able to distinguish it from normal functions.
 
-                let is_variadic = parameters.iter().any(|p| p.2);
-
-                self.vm.add_foreign_function(
-                    name,
-                    if is_variadic {
-                        1..=usize::MAX
-                    } else {
-                        parameters.len()..=parameters.len()
-                    },
-                );
+                self.vm
+                    .add_foreign_function(name, parameters.len()..=parameters.len());
 
                 self.functions.insert(name.clone(), true);
             }
@@ -372,7 +403,7 @@ impl BytecodeInterpreter {
                 // Declaring a dimension is like introducing a new type. The information
                 // is only relevant for the type checker. Nothing happens at run time.
             }
-            Statement::DefineBaseUnit(unit_name, decorators, readable_type, type_) => {
+            Statement::DefineBaseUnit(unit_name, decorators, annotation, type_) => {
                 let aliases = decorator::name_and_aliases(unit_name, decorators)
                     .map(|(name, ap)| (name.clone(), ap))
                     .collect();
@@ -382,8 +413,11 @@ impl BytecodeInterpreter {
                     .add_base_unit(
                         unit_name,
                         UnitMetadata {
-                            type_: type_.clone(),
-                            readable_type: readable_type.clone(),
+                            type_: type_.to_concrete_type(), // Base unit types can never be generic
+                            readable_type: annotation
+                                .as_ref()
+                                .map(|a| a.pretty_print())
+                                .unwrap_or(type_.to_readable_type(dimension_registry)),
                             aliases,
                             name: decorator::name(decorators),
                             canonical_name: decorator::get_canonical_unit_name(
@@ -406,7 +440,14 @@ impl BytecodeInterpreter {
                         .insert(name.into(), constant_idx);
                 }
             }
-            Statement::DefineDerivedUnit(unit_name, expr, decorators, readable_type, type_) => {
+            Statement::DefineDerivedUnit(
+                unit_name,
+                expr,
+                decorators,
+                annotation,
+                type_,
+                _readable_type,
+            ) => {
                 let aliases = decorator::name_and_aliases(unit_name, decorators)
                     .map(|(name, ap)| (name.clone(), ap))
                     .collect();
@@ -428,8 +469,11 @@ impl BytecodeInterpreter {
                         .name,
                     ),
                     UnitMetadata {
-                        type_: type_.clone(),
-                        readable_type: readable_type.clone(),
+                        type_: type_.to_concrete_type(), // We guarantee that derived-unit definitions do not contain generics, so no TGen(..)s can escape
+                        readable_type: annotation
+                            .as_ref()
+                            .map(|a| a.pretty_print())
+                            .unwrap_or(type_.to_readable_type(dimension_registry)),
                         aliases,
                         name: decorator::name(decorators),
                         canonical_name: decorator::get_canonical_unit_name(unit_name, decorators),
@@ -456,7 +500,7 @@ impl BytecodeInterpreter {
 
                 use crate::markup as m;
                 let idx = self.vm.add_string(
-                    m::dimmed("=") + m::whitespace(" ") + arg.get_type().pretty_print(),
+                    m::dimmed("=") + m::whitespace(" ") + arg.get_type_scheme().pretty_print(), // TODO
                 );
                 self.vm.add_op1(Op::PrintString, idx);
             }
@@ -468,9 +512,18 @@ impl BytecodeInterpreter {
 
                 let name = &ffi::procedures().get(kind).unwrap().name;
 
-                let idx = self.vm.get_ffi_callable_idx(name).unwrap();
-                self.vm
-                    .add_op2(Op::FFICallProcedure, idx, args.len() as u16); // TODO: check overflow
+                let callable_idx = self.vm.get_ffi_callable_idx(name).unwrap();
+
+                let arg_spans = args.iter().map(|a| a.full_span()).collect();
+                let spans_idx = self.vm.add_procedure_arg_span(arg_spans);
+
+                self.vm.add_op3(
+                    Op::FFICallProcedure,
+                    callable_idx,
+                    args.len() as u16,
+                    spans_idx,
+                );
+                // TODO: check overflow
             }
             Statement::DefineStruct(struct_info) => {
                 self.vm.add_struct_info(struct_info);
@@ -531,9 +584,10 @@ impl Interpreter for BytecodeInterpreter {
         &mut self,
         settings: &mut InterpreterSettings,
         statements: &[Statement],
+        dimension_registry: &DimensionRegistry,
     ) -> Result<InterpreterResult> {
         for statement in statements {
-            self.compile_statement(statement)?;
+            self.compile_statement(statement, dimension_registry)?;
         }
 
         self.run(settings)
