@@ -205,28 +205,6 @@ impl ErrorDiagnostic for TypeCheckError {
             TypeCheckError::TypeParameterNameClash(span, _) => d.with_labels(vec![span
                 .diagnostic_label(LabelStyle::Primary)
                 .with_message(inner_error)]),
-            TypeCheckError::CanNotInferTypeParameters(
-                span,
-                callable_definition_span,
-                _,
-                params,
-            ) => d.with_labels(vec![
-                callable_definition_span
-                    .diagnostic_label(LabelStyle::Secondary)
-                    .with_message(format!(
-                        "The type parameter(s) {params} in this generic function"
-                    )),
-                span.diagnostic_label(LabelStyle::Primary)
-                    .with_message("… could not be inferred for this function call"),
-            ]),
-            TypeCheckError::MultipleUnresolvedTypeParameters(span, parameter_span) => d
-                .with_labels(vec![
-                    span.diagnostic_label(LabelStyle::Secondary)
-                        .with_message("In this function call"),
-                    parameter_span
-                        .diagnostic_label(LabelStyle::Primary)
-                        .with_message(inner_error),
-                ]),
             TypeCheckError::IncompatibleTypesInCondition(
                 if_span,
                 then_type,
@@ -259,7 +237,7 @@ impl ErrorDiagnostic for TypeCheckError {
                     .with_message(rhs_type.to_string()),
                 op_span
                     .diagnostic_label(LabelStyle::Primary)
-                    .with_message("Incompatible types comparison operator"),
+                    .with_message("Incompatible types in comparison operator"),
             ]),
             TypeCheckError::IncompatibleTypeInAssert(procedure_span, type_, type_span) => d
                 .with_labels(vec![
@@ -328,6 +306,21 @@ impl ErrorDiagnostic for TypeCheckError {
                         .with_notes(vec![inner_error])
                 }
             }
+            TypeCheckError::IncompatibleTypesInList(
+                span_first,
+                type_first,
+                span_subsequent,
+                type_subsequent,
+            ) => d
+                .with_labels(vec![
+                    span_first
+                        .diagnostic_label(LabelStyle::Secondary)
+                        .with_message(type_first.to_string()),
+                    span_subsequent
+                        .diagnostic_label(LabelStyle::Primary)
+                        .with_message(type_subsequent.to_string()),
+                ])
+                .with_notes(vec![inner_error]),
             TypeCheckError::NoDimensionlessBaseUnit(span, unit_name) => d
                 .with_labels(vec![span
                     .diagnostic_label(LabelStyle::Primary)
@@ -343,10 +336,11 @@ impl ErrorDiagnostic for TypeCheckError {
             | TypeCheckError::ExpectedDimensionType(span, _)
             | TypeCheckError::ExpectedBool(span)
             | TypeCheckError::NoFunctionReferenceToGenericFunction(span)
-            | TypeCheckError::OnlyFunctionsAndReferencesCanBeCalled(span) => d
-                .with_labels(vec![span
-                    .diagnostic_label(LabelStyle::Primary)
-                    .with_message(inner_error)]),
+            | TypeCheckError::OnlyFunctionsAndReferencesCanBeCalled(span)
+            | TypeCheckError::DerivedUnitDefinitionMustNotBeGeneric(span)
+            | TypeCheckError::MultipleTypedHoles(span) => d.with_labels(vec![span
+                .diagnostic_label(LabelStyle::Primary)
+                .with_message(inner_error)]),
             TypeCheckError::MissingDimension(span, dim) => d
                 .with_labels(vec![span
                     .diagnostic_label(LabelStyle::Primary)
@@ -464,6 +458,38 @@ impl ErrorDiagnostic for TypeCheckError {
             TypeCheckError::NameResolutionError(inner) => {
                 return inner.diagnostics();
             }
+            TypeCheckError::ConstraintSolverError(..) | TypeCheckError::SubstitutionError(..) => {
+                d.with_message(inner_error).with_notes(vec![
+                    "Consider adding type annotations to get more precise error messages.".into(),
+                ])
+            }
+            TypeCheckError::MissingDimBound(span) => d
+                .with_labels(vec![span
+                    .diagnostic_label(LabelStyle::Primary)
+                    .with_message(inner_error)])
+                .with_notes(vec![
+                    "Consider adding `: Dim` after the type parameter".to_owned()
+                ]),
+            TypeCheckError::ExponentiationNeedsTypeAnnotation(span) => d.with_labels(vec![span
+                .diagnostic_label(LabelStyle::Primary)
+                .with_message(inner_error)]),
+            TypeCheckError::TypedHoleInStatement(span, type_, statement, matches) => {
+                let mut notes = vec![
+                    format!("Found a hole of type '{type_}' in the statement:"),
+                    format!("  {statement}"),
+                ];
+
+                if !matches.is_empty() {
+                    notes.push("Relevant matches for this hole include:".into());
+                    notes.push(format!("  {}", matches.join(", ")));
+                }
+
+                d.with_labels(vec![span
+                    .diagnostic_label(LabelStyle::Primary)
+                    .with_message(type_)])
+                    .with_message("Found typed hole")
+                    .with_notes(notes)
+            }
         };
         vec![d]
     }
@@ -471,8 +497,43 @@ impl ErrorDiagnostic for TypeCheckError {
 
 impl ErrorDiagnostic for RuntimeError {
     fn diagnostics(&self) -> Vec<Diagnostic> {
-        vec![Diagnostic::error()
-            .with_message("runtime error")
-            .with_notes(vec![format!("{self:#}")])]
+        let inner = format!("{self:#}");
+
+        match self {
+            RuntimeError::AssertFailed(span) => vec![Diagnostic::error()
+                .with_message("assertion failed")
+                .with_labels(vec![span
+                    .diagnostic_label(LabelStyle::Primary)
+                    .with_message("assertion failed")])],
+            RuntimeError::AssertEq2Failed(span_lhs, lhs, span_rhs, rhs) => {
+                vec![Diagnostic::error()
+                    .with_message("Assertion failed")
+                    .with_labels(vec![
+                        span_lhs
+                            .diagnostic_label(LabelStyle::Secondary)
+                            .with_message(format!("{lhs}")),
+                        span_rhs
+                            .diagnostic_label(LabelStyle::Primary)
+                            .with_message(format!("{rhs}")),
+                    ])
+                    .with_notes(vec![inner])]
+            }
+            RuntimeError::AssertEq3Failed(span_lhs, lhs, span_rhs, rhs, _) => {
+                vec![Diagnostic::error()
+                    .with_message("Assertion failed")
+                    .with_labels(vec![
+                        span_lhs
+                            .diagnostic_label(LabelStyle::Secondary)
+                            .with_message(format!("{lhs}")),
+                        span_rhs
+                            .diagnostic_label(LabelStyle::Primary)
+                            .with_message(format!("{rhs}")),
+                    ])
+                    .with_notes(vec![format!("{self:#}")])]
+            }
+            _ => vec![Diagnostic::error()
+                .with_message("runtime error")
+                .with_notes(vec![inner])],
+        }
     }
 }
