@@ -6,6 +6,7 @@ pub const KEYWORDS: &[&str] = &[
     "to ",
     "let ",
     "fn ",
+    "where ",
     "dimension ",
     "unit ",
     "use ",
@@ -31,6 +32,8 @@ pub const KEYWORDS: &[&str] = &[
     "Bool",
     "String",
     "DateTime",
+    "Fn",
+    "List",
     // decorators
     "metric_prefixes",
     "binary_prefixes",
