@@ -16,6 +16,7 @@ pub mod help;
 pub mod html_formatter;
 mod interpreter;
 pub mod keywords;
+pub mod list;
 pub mod markup;
 mod math;
 pub mod module_importer;
@@ -33,6 +34,8 @@ pub mod resolver;
 mod span;
 mod suggestion;
 mod tokenizer;
+mod traversal;
+mod type_variable;
 mod typechecker;
 mod typed_ast;
 pub mod unicode_input;
@@ -47,7 +50,6 @@ use currency::ExchangeRatesCache;
 use diagnostic::ErrorDiagnostic;
 use dimension::DimensionRegistry;
 use interpreter::Interpreter;
-use itertools::Itertools;
 use keywords::KEYWORDS;
 use markup as m;
 use markup::FormatType;
@@ -150,6 +152,38 @@ impl Context {
             .cloned()
     }
 
+    pub fn functions(
+        &self,
+    ) -> impl Iterator<
+        Item = (
+            String,
+            Option<String>,
+            String,
+            Option<String>,
+            Option<String>,
+            CodeSource,
+        ),
+    > + '_ {
+        self.prefix_transformer
+            .function_names
+            .iter()
+            .filter(|name| !name.starts_with('_'))
+            .map(move |name| {
+                let (signature, meta) = self.typechecker.lookup_function(name).unwrap();
+                (
+                    name.clone(),
+                    meta.name.clone(),
+                    signature
+                        .pretty_print(self.dimension_registry())
+                        .to_string(),
+                    meta.description.clone(),
+                    meta.url.clone(),
+                    self.resolver
+                        .get_code_source(signature.definition_span.code_source_id),
+                )
+            })
+    }
+
     pub fn unit_names(&self) -> &[Vec<String>] {
         &self.prefix_transformer.unit_names
     }
@@ -331,7 +365,7 @@ impl Context {
                         + m::text("])")
                         + m::nl();
                 } else {
-                    help += m::text("A unit of [") + md.readable_type + m::text("]") + m::nl();
+                    help += m::text("A unit of: ") + md.readable_type + m::nl();
                 }
 
                 if let Some(defining_info) = self.interpreter.get_defining_unit(&full_name) {
@@ -421,21 +455,8 @@ impl Context {
 
         if let Some((fn_signature, fn_metadata)) = self.typechecker.lookup_function(keyword) {
             let metadata = fn_metadata.clone();
-            let parameters = fn_signature
-                .parameter_types
-                .iter()
-                .map(|(_, name, typ)| (name, typ.to_readable_type(self.dimension_registry())))
-                .collect_vec();
-            let type_parameters = fn_signature
-                .type_parameters
-                .iter()
-                .map(|(_, s)| s)
-                .collect_vec();
-            let return_type = fn_signature
-                .return_type
-                .to_readable_type(self.dimension_registry());
 
-            let mut help = m::text("Function: ");
+            let mut help = m::text("Function:    ");
             if let Some(name) = &metadata.name {
                 help += m::text(name);
             } else {
@@ -446,29 +467,10 @@ impl Context {
             }
             help += m::nl();
 
-            help += m::text("Signature:") + m::space() + m::identifier(keyword);
-            if !type_parameters.is_empty() {
-                help += m::operator("<");
-                help += Itertools::intersperse(
-                    type_parameters.iter().map(|t| m::type_identifier(t)),
-                    m::operator(",") + m::space(),
-                )
-                .sum();
-                help += m::operator(">");
-            }
-            help += m::operator("(");
-            help += Itertools::intersperse(
-                parameters
-                    .iter()
-                    .map(|(p, t)| m::identifier(p) + m::text(":") + m::space() + t.clone()),
-                m::operator(",") + m::space(),
-            )
-            .sum();
-            if fn_signature.is_variadic {
-                help += m::operator("…");
-            }
-            help += m::operator(")");
-            help += m::space() + m::operator("->") + m::space() + return_type + m::nl();
+            help += m::text("Signature:  ")
+                + m::space()
+                + fn_signature.pretty_print(self.typechecker.registry())
+                + m::nl();
 
             if let Some(description) = &metadata.description {
                 let desc = "Description: ";
@@ -571,7 +573,7 @@ impl Context {
 
         let result = self
             .typechecker
-            .check_statements(transformed_statements)
+            .check(transformed_statements)
             .map_err(NumbatError::TypeCheckError);
 
         if result.is_err() {
@@ -754,9 +756,13 @@ impl Context {
 
         let typed_statements = result?;
 
-        let result = self
-            .interpreter
-            .interpret_statements(settings, &typed_statements);
+        let interpreter_old = self.interpreter.clone();
+
+        let result = self.interpreter.interpret_statements(
+            settings,
+            &typed_statements,
+            self.typechecker.registry(),
+        );
 
         if result.is_err() {
             // Similar to above: we need to reset the state of the typechecker and the prefix transformer
@@ -771,6 +777,7 @@ impl Context {
             //
             self.prefix_transformer = prefix_transformer_old;
             self.typechecker = typechecker_old;
+            self.interpreter = interpreter_old;
         }
 
         let result = result.map_err(NumbatError::RuntimeError)?;
