@@ -63,7 +63,6 @@ pub enum TokenKind {
     Multiply,
     Power,
     Divide,
-    Per,
     Comma,
     Arrow,
     Equal,
@@ -83,37 +82,41 @@ pub enum TokenKind {
     LogicalAnd,
     LogicalOr,
     Period,
+    QuestionMark,
 
     // Keywords
+    Per,
+    To,
     Let,
     Fn, // 'fn'
+    Where,
+    And,
     Dimension,
     Unit,
     Use,
     Struct,
 
-    To,
-
-    Bool,
-    True,
-    False,
-    If,
-    Then,
-    Else,
-
-    String,
-    DateTime,
-
-    NaN,
-    Inf,
-
-    CapitalFn, // 'Fn'
-
     Long,
     Short,
     Both,
     None,
 
+    If,
+    Then,
+    Else,
+    True,
+    False,
+
+    NaN,
+    Inf,
+
+    // Type names
+    Bool,
+    String,
+    DateTime,
+    CapitalFn, // 'Fn'
+    List,
+
     // Procedure calls
     ProcedurePrint,
     ProcedureAssert,
@@ -199,6 +202,8 @@ fn is_identifier_start(c: char) -> bool {
         || is_currency_char(c)
         || is_other_allowed_identifier_char(c)
         || c == '°'
+        || c == '′'
+        || c == '″'
         || c == '_'
 }
 
@@ -342,37 +347,66 @@ impl Tokenizer {
         Ok(())
     }
 
+    fn consume_string(&mut self) -> Result<()> {
+        let mut escaped = false;
+        loop {
+            escaped = match self.peek() {
+                None => {
+                    break;
+                }
+                Some('\\') if !escaped => true,
+                Some('"') | Some('{') if !escaped => {
+                    break;
+                }
+                Some(_) => false,
+            };
+
+            self.advance();
+        }
+
+        Ok(())
+    }
+
     fn scan_single_token(&mut self) -> Result<Option<Token>> {
         static KEYWORDS: OnceLock<HashMap<&'static str, TokenKind>> = OnceLock::new();
         let keywords = KEYWORDS.get_or_init(|| {
             let mut m = HashMap::new();
+            // keywords
             m.insert("per", TokenKind::Per);
             m.insert("to", TokenKind::To);
             m.insert("let", TokenKind::Let);
             m.insert("fn", TokenKind::Fn);
-            m.insert("struct", TokenKind::Struct);
+            m.insert("where", TokenKind::Where);
+            m.insert("and", TokenKind::And);
             m.insert("dimension", TokenKind::Dimension);
             m.insert("unit", TokenKind::Unit);
             m.insert("use", TokenKind::Use);
+            m.insert("struct", TokenKind::Struct);
             m.insert("long", TokenKind::Long);
             m.insert("short", TokenKind::Short);
             m.insert("both", TokenKind::Both);
             m.insert("none", TokenKind::None);
+            m.insert("if", TokenKind::If);
+            m.insert("then", TokenKind::Then);
+            m.insert("else", TokenKind::Else);
+            m.insert("true", TokenKind::True);
+            m.insert("false", TokenKind::False);
+            m.insert("NaN", TokenKind::NaN);
+            m.insert("inf", TokenKind::Inf);
+
+            // procedures
             m.insert("print", TokenKind::ProcedurePrint);
             m.insert("assert", TokenKind::ProcedureAssert);
             m.insert("assert_eq", TokenKind::ProcedureAssertEq);
             m.insert("type", TokenKind::ProcedureType);
+
+            // type names
             m.insert("Bool", TokenKind::Bool);
-            m.insert("true", TokenKind::True);
-            m.insert("false", TokenKind::False);
-            m.insert("if", TokenKind::If);
-            m.insert("then", TokenKind::Then);
-            m.insert("else", TokenKind::Else);
             m.insert("String", TokenKind::String);
             m.insert("DateTime", TokenKind::DateTime);
             m.insert("Fn", TokenKind::CapitalFn);
-            m.insert("NaN", TokenKind::NaN);
-            m.insert("inf", TokenKind::Inf);
+            m.insert("List", TokenKind::List);
+
             // Keep this list in sync with keywords::KEYWORDS!
             m
         });
@@ -413,6 +447,7 @@ impl Tokenizer {
             '≥' => TokenKind::GreaterOrEqual,
             '>' if self.match_char('=') => TokenKind::GreaterOrEqual,
             '>' => TokenKind::GreaterThan,
+            '?' => TokenKind::QuestionMark,
             '0' if self
                 .peek()
                 .map(|c| c == 'x' || c == 'o' || c == 'b')
@@ -499,10 +534,10 @@ impl Tokenizer {
             '\n' => TokenKind::Newline,
             '&' if self.match_char('&') => TokenKind::LogicalAnd,
             '|' if self.match_char('|') => TokenKind::LogicalOr,
+            '|' if self.match_char('>') => TokenKind::PostfixApply,
             '*' if self.match_char('*') => TokenKind::Power,
             '+' => TokenKind::Plus,
             '*' | '·' | '×' => TokenKind::Multiply,
-            '/' if self.match_char('/') => TokenKind::PostfixApply,
             '/' => TokenKind::Divide,
             '÷' => TokenKind::Divide,
             '^' => TokenKind::Power,
@@ -536,9 +571,7 @@ impl Tokenizer {
                 InterpolationState::Outside => {
                     self.string_start = self.token_start;
 
-                    while self.peek().map(|c| c != '"' && c != '{').unwrap_or(false) {
-                        self.advance();
-                    }
+                    self.consume_string()?;
 
                     if self.match_char('"') {
                         TokenKind::StringFixed
@@ -597,9 +630,7 @@ impl Tokenizer {
                 }
             }
             '}' if self.interpolation_state.is_inside() => {
-                while self.peek().map(|c| c != '"' && c != '{').unwrap_or(false) {
-                    self.advance();
-                }
+                self.consume_string()?;
 
                 if self.match_char('"') {
                     self.interpolation_state = InterpolationState::Outside;
@@ -1091,6 +1122,32 @@ fn test_tokenize_string() {
         tokenize("\"foo = {foo, bar = {bar}\"", 0).unwrap_err().kind,
         TokenizerErrorKind::UnexpectedCurlyInInterpolation
     );
+
+    insta::assert_snapshot!(
+        tokenize_reduced_pretty(r#""start \"inner\" end""#).unwrap(),
+        @r###"
+    "\"start \\\"inner\\\" end\"", StringFixed, (1, 1)
+    "", Eof, (1, 22)
+    "###
+    );
+
+    insta::assert_snapshot!(
+        tokenize_reduced_pretty(r#""start \{inner\} end""#).unwrap(),
+        @r###"
+    "\"start \\{inner\\} end\"", StringFixed, (1, 1)
+    "", Eof, (1, 22)
+    "###
+    );
+
+    insta::assert_snapshot!(
+        tokenize_reduced_pretty(r#""start {1} \"inner\" end""#).unwrap(),
+        @r###"
+    "\"start {", StringInterpolationStart, (1, 1)
+    "1", Number, (1, 9)
+    "} \\\"inner\\\" end\"", StringInterpolationEnd, (1, 10)
+    "", Eof, (1, 26)
+    "###
+    );
 }
 
 #[test]
@@ -1186,3 +1243,20 @@ fn test_field_access() {
         @"Error at (1, 11): `Expected digit`"
     );
 }
+
+#[test]
+fn test_lists() {
+    insta::assert_snapshot!(
+        tokenize_reduced_pretty("[1, 2.3, 4]").unwrap(),
+        @r###"
+    "[", LeftBracket, (1, 1)
+    "1", Number, (1, 2)
+    ",", Comma, (1, 3)
+    "2.3", Number, (1, 5)
+    ",", Comma, (1, 8)
+    "4", Number, (1, 10)
+    "]", RightBracket, (1, 11)
+    "", Eof, (1, 12)
+    "###
+    );
+}
