@@ -1,9 +1,12 @@
-use std::collections::HashMap;
+use std::collections::{HashMap, VecDeque};
 use std::sync::Arc;
 use std::{cmp::Ordering, fmt::Display};
 
 use indexmap::IndexMap;
+use num_traits::ToPrimitive;
 
+use crate::list::NumbatList;
+use crate::span::Span;
 use crate::typed_ast::StructInfo;
 use crate::{
     ffi::{self, ArityRange, Callable, ForeignFunction},
@@ -93,6 +96,7 @@ pub enum Op {
     /// Same as above, but call a foreign/native function
     FFICallFunction,
     /// Same as above, but call a procedure which does not return anything (does not push a value onto the stack)
+    /// It has a third argument which is an index to retrieve the source-span of the arguments
     FFICallProcedure,
 
     /// Call a callable object
@@ -109,10 +113,12 @@ pub enum Op {
 
     /// Build a struct from the field values on the stack
     BuildStructInstance,
-
     /// Access a single field of a struct
     AccessStructField,
 
+    /// Build a list from the elements on the stack
+    BuildList,
+
     /// Return from the current function
     Return,
 }
@@ -120,11 +126,8 @@ pub enum Op {
 impl Op {
     fn num_operands(self) -> usize {
         match self {
-            Op::SetUnitConstant
-            | Op::Call
-            | Op::FFICallFunction
-            | Op::FFICallProcedure
-            | Op::BuildStructInstance => 2,
+            Op::FFICallProcedure => 3,
+            Op::SetUnitConstant | Op::Call | Op::FFICallFunction | Op::BuildStructInstance => 2,
             Op::LoadConstant
             | Op::ApplyPrefix
             | Op::GetLocal
@@ -134,7 +137,8 @@ impl Op {
             | Op::JumpIfFalse
             | Op::Jump
             | Op::CallCallable
-            | Op::AccessStructField => 1,
+            | Op::AccessStructField
+            | Op::BuildList => 1,
             Op::Negate
             | Op::Factorial
             | Op::Add
@@ -201,6 +205,7 @@ impl Op {
             Op::Return => "Return",
             Op::BuildStructInstance => "BuildStructInstance",
             Op::AccessStructField => "AccessStructField",
+            Op::BuildList => "BuildList",
         }
     }
 }
@@ -303,6 +308,10 @@ pub struct Vm {
     /// List of registered native/foreign functions
     ffi_callables: Vec<&'static ForeignFunction>,
 
+    /// Spans for arguments of procedure calls. This is used for
+    /// assertion error messages, for example.
+    procedure_arg_spans: Vec<Vec<Span>>,
+
     /// The call stack
     frames: Vec<CallFrame>,
 
@@ -327,6 +336,7 @@ impl Vm {
             unit_information: vec![],
             last_result: None,
             ffi_callables: ffi::procedures().iter().map(|(_, ff)| ff).collect(),
+            procedure_arg_spans: vec![],
             frames: vec![CallFrame::root()],
             stack: vec![],
             debug: false,
@@ -366,6 +376,14 @@ impl Vm {
         Self::push_u16(current_chunk, arg2);
     }
 
+    pub(crate) fn add_op3(&mut self, op: Op, arg1: u16, arg2: u16, arg3: u16) {
+        let current_chunk = self.current_chunk_mut();
+        current_chunk.push(op as u8);
+        Self::push_u16(current_chunk, arg1);
+        Self::push_u16(current_chunk, arg2);
+        Self::push_u16(current_chunk, arg3);
+    }
+
     pub fn current_offset(&self) -> u16 {
         self.bytecode[self.current_chunk_index].1.len() as u16
     }
@@ -461,6 +479,12 @@ impl Vm {
         Some(position as u16)
     }
 
+    pub(crate) fn add_procedure_arg_span(&mut self, spans: Vec<Span>) -> u16 {
+        self.procedure_arg_spans.push(spans);
+        assert!(self.procedure_arg_spans.len() <= u16::MAX as usize);
+        (self.procedure_arg_spans.len() - 1) as u16
+    }
+
     pub fn disassemble(&self) {
         if !self.debug {
             return;
@@ -567,7 +591,7 @@ impl Vm {
     }
 
     #[track_caller]
-    fn pop_datetime(&mut self) -> chrono::DateTime<chrono::FixedOffset> {
+    fn pop_datetime(&mut self) -> jiff::Zoned {
         match self.pop() {
             Value::DateTime(q) => q,
             _ => panic!("Expected datetime to be on the top of the stack"),
@@ -690,21 +714,24 @@ impl Vm {
 
                     // for time, the base unit is in seconds
                     let base = rhs.to_base_unit_representation();
-                    let seconds_f = base.unsafe_value().to_f64();
+                    let seconds_f64 = base.unsafe_value().to_f64();
 
-                    let duration = chrono::Duration::try_seconds(seconds_f.trunc() as i64)
-                        .ok_or(RuntimeError::DurationOutOfRange)?
-                        + chrono::Duration::nanoseconds(
-                            (seconds_f.fract() * 1_000_000_000f64).round() as i64,
-                        );
+                    let seconds_i64 = seconds_f64
+                        .to_i64()
+                        .ok_or(RuntimeError::DurationOutOfRange)?;
+
+                    let span = jiff::Span::new()
+                        .try_seconds(seconds_i64)
+                        .map_err(|_| RuntimeError::DurationOutOfRange)?
+                        .nanoseconds((seconds_f64.fract() * 1_000_000_000f64).round() as i64);
 
                     self.push(Value::DateTime(match op {
                         Op::AddToDateTime => lhs
-                            .checked_add_signed(duration)
-                            .ok_or(RuntimeError::DateTimeOutOfRange)?,
+                            .checked_add(span)
+                            .map_err(|_| RuntimeError::DateTimeOutOfRange)?,
                         Op::SubFromDateTime => lhs
-                            .checked_sub_signed(duration)
-                            .ok_or(RuntimeError::DateTimeOutOfRange)?,
+                            .checked_sub(span)
+                            .map_err(|_| RuntimeError::DateTimeOutOfRange)?,
                         _ => unreachable!(),
                     }));
                 }
@@ -713,9 +740,12 @@ impl Vm {
                     let rhs = self.pop_datetime();
                     let lhs = self.pop_datetime();
 
-                    let duration = lhs - rhs;
-                    let duration = duration.subsec_nanos() as f64 / 1_000_000_000f64
-                        + duration.num_seconds() as f64;
+                    let duration = lhs
+                        .since(&rhs)
+                        .map_err(|_| RuntimeError::DateTimeOutOfRange)?;
+                    let duration = duration
+                        .total(jiff::Unit::Second)
+                        .map_err(|_| RuntimeError::DurationOutOfRange)?;
 
                     let ret = Value::Quantity(Quantity::new(
                         Number::from_f64(duration),
@@ -816,19 +846,21 @@ impl Vm {
 
                     debug_assert!(foreign_function.arity.contains(&num_args));
 
-                    let mut args = vec![];
+                    let mut args = VecDeque::new();
                     for _ in 0..num_args {
-                        args.push(self.pop());
+                        args.push_front(self.pop());
                     }
-                    args.reverse(); // TODO: use a deque?
 
                     match &self.ffi_callables[function_idx].callable {
                         Callable::Function(function) => {
-                            let result = (function)(&args[..]);
+                            let result = (function)(args);
                             self.push(result?);
                         }
                         Callable::Procedure(procedure) => {
-                            let result = (procedure)(ctx, &args[..]);
+                            let span_idx = self.read_u16() as usize;
+                            let spans = &self.procedure_arg_spans[span_idx];
+
+                            let result = (procedure)(ctx, args, spans.clone());
 
                             match result {
                                 std::ops::ControlFlow::Continue(()) => {}
@@ -844,7 +876,7 @@ impl Vm {
 
                     let callable = self.pop();
                     match callable.unsafe_as_function_reference() {
-                        FunctionReference::Normal(name) => {
+                        FunctionReference::Normal(ref name) => {
                             let function_idx = self.get_function_idx(name) as usize;
 
                             // TODO: unify code with 'Op::Call'?
@@ -854,21 +886,20 @@ impl Vm {
                                 fp: self.stack.len() - num_args,
                             })
                         }
-                        FunctionReference::Foreign(name) => {
+                        FunctionReference::Foreign(ref name) => {
                             let function_idx = self
                                 .get_ffi_callable_idx(name)
                                 .expect("Foreign function exists")
                                 as usize;
 
-                            let mut args = vec![];
+                            let mut args = VecDeque::new();
                             for _ in 0..num_args {
-                                args.push(self.pop());
+                                args.push_front(self.pop());
                             }
-                            args.reverse();
 
                             match &self.ffi_callables[function_idx].callable {
                                 Callable::Function(function) => {
-                                    let result = (function)(&args[..]);
+                                    let result = (function)(args);
                                     self.push(result?);
                                 }
                                 Callable::Procedure(..) => unreachable!("Foreign procedures can not be targeted by a function reference"),
@@ -879,11 +910,10 @@ impl Vm {
 
                             let dt = self.pop_datetime();
 
-                            let tz: chrono_tz::Tz = tz_name
-                                .parse()
+                            let tz = jiff::tz::TimeZone::get(&tz_name)
                                 .map_err(|_| RuntimeError::UnknownTimezone(tz_name.into()))?;
 
-                            let dt = dt.with_timezone(&tz).fixed_offset();
+                            let dt = dt.with_time_zone(tz);
 
                             self.push(Value::DateTime(dt));
                         }
@@ -901,9 +931,10 @@ impl Vm {
                         Value::Quantity(q) => q.to_string(),
                         Value::Boolean(b) => b.to_string(),
                         Value::String(s) => s,
-                        Value::DateTime(dt) => crate::datetime::to_rfc2822_save(&dt),
+                        Value::DateTime(dt) => crate::datetime::to_string(&dt),
                         Value::FunctionReference(r) => r.to_string(),
                         s @ Value::StructInstance(..) => s.to_string(),
+                        l @ Value::List(_) => l.to_string(),
                         Value::FormatSpecifiers(_) => unreachable!(),
                     };
 
@@ -1004,6 +1035,16 @@ impl Vm {
                     let value = fields.swap_remove(field_idx as usize);
                     self.stack.push(value);
                 }
+                Op::BuildList => {
+                    let length = self.read_u16();
+                    let mut list = NumbatList::with_capacity(length as usize);
+
+                    for _ in 0..length {
+                        list.push_front(self.pop());
+                    }
+
+                    self.stack.push(list.into());
+                }
             }
         }
 
