@@ -47,8 +47,22 @@ fn expect_output(code: &str, expected_output: impl AsRef<str>) {
 }
 
 #[track_caller]
-fn expect_failure(code: &str, msg_part: &str) {
+fn expect_pretty_print(code: &str, expected_pretty_print_output: impl AsRef<str>) {
     let mut ctx = get_test_context();
+
+    let (statements, _result) = ctx.interpret(code, CodeSource::Internal).unwrap();
+
+    assert_eq!(statements.len(), 1);
+
+    let statement = &statements[0];
+    assert_eq!(
+        statement.pretty_print().to_string(),
+        expected_pretty_print_output.as_ref()
+    );
+}
+
+#[track_caller]
+fn expect_failure_with_context(ctx: &mut Context, code: &str, msg_part: &str) {
     if let Err(e) = ctx.interpret(code, CodeSource::Internal) {
         let error_message = e.to_string();
         println!("{}", error_message);
@@ -61,6 +75,12 @@ fn expect_failure(code: &str, msg_part: &str) {
     }
 }
 
+#[track_caller]
+fn expect_failure(code: &str, msg_part: &str) {
+    let mut ctx = get_test_context();
+    expect_failure_with_context(&mut ctx, code, msg_part)
+}
+
 #[track_caller]
 fn get_error_message(code: &str) -> String {
     let mut ctx = get_test_context();
@@ -236,23 +256,19 @@ fn test_algebra() {
     let _ = ctx
         .interpret("use extra::algebra", CodeSource::Internal)
         .unwrap();
-    expect_output_with_context(&mut ctx, "quadratic_equation(1, 0, -1)", "x₁ = 1; x₂ = -1");
-    expect_output_with_context(&mut ctx, "quadratic_equation(0, 9, 3)", "x = -0.333333");
-    expect_output_with_context(&mut ctx, "quadratic_equation(0, 0, 1)", "no solution");
-    expect_output_with_context(&mut ctx, "quadratic_equation(9, -126, 441)", "x = 7");
-    expect_output_with_context(&mut ctx, "quadratic_equation(1, -2, 1)", "x = 1");
-    expect_output_with_context(&mut ctx, "quadratic_equation(0, 1, 1)", "x = -1");
-    expect_output_with_context(&mut ctx, "quadratic_equation(1, 0, 0)", "x = 0");
-    expect_output_with_context(
+    expect_output_with_context(&mut ctx, "quadratic_equation(1, 0, -1)", "[1, -1]");
+    expect_output_with_context(&mut ctx, "quadratic_equation(0, 9, 3)", "[-0.333333]");
+    expect_output_with_context(&mut ctx, "quadratic_equation(0, 0, 1)", "[]");
+    expect_output_with_context(&mut ctx, "quadratic_equation(9, -126, 441)", "[7]");
+    expect_output_with_context(&mut ctx, "quadratic_equation(1, -2, 1)", "[1]");
+    expect_output_with_context(&mut ctx, "quadratic_equation(0, 1, 1)", "[-1]");
+    expect_output_with_context(&mut ctx, "quadratic_equation(1, 0, 0)", "[0]");
+    expect_failure_with_context(
         &mut ctx,
         "quadratic_equation(0, 0, 0)",
         "infinitely many solutions",
     );
-    expect_output_with_context(
-        &mut ctx,
-        "quadratic_equation(1, 1, 1)",
-        "no real-valued solution",
-    );
+    expect_output_with_context(&mut ctx, "quadratic_equation(1, 1, 1)", "[]");
 }
 
 #[test]
@@ -265,7 +281,7 @@ fn test_math() {
     expect_output("atan2(100 cm, 1 m) / (pi / 4)", "1");
     expect_failure(
         "atan2(100 cm, 1 m²)",
-        "parameter type: Length\n argument type: Length²",
+        "Could not solve the following constraints",
     );
 
     expect_output("mod(5, 3)", "2");
@@ -273,10 +289,7 @@ fn test_math() {
     expect_output("mod(8 cm, 5 cm)", "3 cm");
     expect_output("mod(235 cm, 1 m)", "35 cm");
     expect_output("mod(2 m, 7 cm)", "0.04 m");
-    expect_failure(
-        "mod(8 m, 5 s)",
-        "parameter type: Length\n argument type: Time",
-    )
+    expect_failure("mod(8 m, 5 s)", "Could not solve the following constraints")
 }
 
 #[test]
@@ -563,14 +576,14 @@ fn test_logical() {
     expect_output("false || true && !false", "true");
 
     // Errors
-    insta::assert_display_snapshot!(fail("1 || 2"), @"Expected boolean value");
-    insta::assert_display_snapshot!(fail("true || 2"), @"Expected boolean value");
-    insta::assert_display_snapshot!(fail("1 || true"), @"Expected boolean value");
-    insta::assert_display_snapshot!(fail("1 && 2"), @"Expected boolean value");
-    insta::assert_display_snapshot!(fail("true && 2"), @"Expected boolean value");
-    insta::assert_display_snapshot!(fail("1 && true"), @"Expected boolean value");
-    insta::assert_display_snapshot!(fail("!1"), @"Expected boolean value");
-    insta::assert_display_snapshot!(fail("!1 || true"), @"Expected boolean value");
+    insta::assert_snapshot!(fail("1 || 2"), @"Expected boolean value");
+    insta::assert_snapshot!(fail("true || 2"), @"Expected boolean value");
+    insta::assert_snapshot!(fail("1 || true"), @"Expected boolean value");
+    insta::assert_snapshot!(fail("1 && 2"), @"Expected boolean value");
+    insta::assert_snapshot!(fail("true && 2"), @"Expected boolean value");
+    insta::assert_snapshot!(fail("1 && true"), @"Expected boolean value");
+    insta::assert_snapshot!(fail("!1"), @"Expected boolean value");
+    insta::assert_snapshot!(fail("!1 || true"), @"Expected boolean value");
 }
 
 #[test]
@@ -579,37 +592,37 @@ fn test_conditionals() {
     expect_output("if 4 < 3 then 2 else 1", "1");
     expect_output(
         "if 4 > 3 then \"four is larger!\" else \"four is not larger!\"",
-        "four is larger!",
+        "\"four is larger!\"",
     );
 }
 
 #[test]
 fn test_string_interpolation() {
-    expect_output("\"pi = {pi}!\"", "pi = 3.14159!");
-    expect_output("\"1 + 2 = {1 + 2}\"", "1 + 2 = 3");
+    expect_output("\"pi = {pi}!\"", "\"pi = 3.14159!\"");
+    expect_output("\"1 + 2 = {1 + 2}\"", "\"1 + 2 = 3\"");
 
-    expect_output("\"{0.2:0.5}\"", "0.20000");
-    expect_output("\"pi ~= {pi:.3}\"", "pi ~= 3.142");
+    expect_output("\"{0.2:0.5}\"", "\"0.20000\"");
+    expect_output("\"pi ~= {pi:.3}\"", "\"pi ~= 3.142\"");
     expect_output(
         "\"both {pi:.3} and {e} are irrational and transcendental numbers\"",
-        "both 3.142 and 2.71828 are irrational and transcendental numbers",
+        "\"both 3.142 and 2.71828 are irrational and transcendental numbers\"",
     );
     expect_output(
         "
         let str = \"1234\"
         \"{str:0.2}\"
         ",
-        "12",
+        "\"12\"",
     );
 
-    expect_output("\"{1_000_300:+.3}\"", "+1000300.000");
+    expect_output("\"{1_000_300:+.3}\"", "\"+1000300.000\"");
 
     expect_output(
         "
         let str = \"1234\"
         \"a {str:^10} b\"
         ",
-        "a    1234    b",
+        "\"a    1234    b\"",
     );
 
     // Doesn't work at the moment, as `strfmt` expects `i64`'s for `#x`, but Numbat deals with `f64`'s
@@ -698,7 +711,7 @@ fn test_datetime_runtime_errors() {
         "Exceeded maximum size for time durations",
     );
     expect_failure(
-        "date(\"2000-01-01\") + 100000000 years",
+        "date(\"2000-01-01\") + 15_000 years",
         "DateTime out of range",
     );
     expect_failure(
@@ -711,9 +724,96 @@ fn test_datetime_runtime_errors() {
 fn test_user_errors() {
     expect_failure("error(\"test\")", "User error: test");
 
-    // Make sure that the never type (!) can be used in all contexts
+    // Make sure that error(…) can be used in all contexts
     expect_failure("- error(\"test\")", "User error: test");
     expect_failure("1 + error(\"test\")", "User error: test");
     expect_failure("1 m + error(\"test\")", "User error: test");
     expect_failure("if 3 < 2 then 2 m else error(\"test\")", "User error: test");
+    expect_failure(
+        "if true then error(\"test\") else \"foo\"",
+        "User error: test",
+    );
+    expect_failure(
+        "if false then \"foo\" else error(\"test\")",
+        "User error: test",
+    );
+}
+
+#[test]
+fn test_recovery_after_runtime_error() {
+    {
+        let mut ctx = get_test_context();
+
+        expect_failure_with_context(&mut ctx, "let x = 1/0", "Division by zero");
+        expect_failure_with_context(&mut ctx, "x", "Unknown identifier 'x'");
+    }
+    {
+        let mut ctx = get_test_context();
+
+        expect_failure_with_context(&mut ctx, "let x = 1/0", "Division by zero");
+        assert!(ctx
+            .interpret("let x = 1", CodeSource::Internal)
+            .unwrap()
+            .1
+            .is_continue());
+        expect_output_with_context(&mut ctx, "x", "1");
+    }
+}
+
+#[test]
+fn test_statement_pretty_printing() {
+    // Let definitions
+    expect_pretty_print("let v = 10 m/s", "let v: Velocity = 10 metre / second");
+    expect_pretty_print(
+        "let v: Length * Frequency = 10 m/s",
+        "let v: Length × Frequency = 10 metre / second",
+    );
+
+    expect_pretty_print("let x = 0 + 1 m", "let x: Length = 0 + 1 metre");
+
+    expect_pretty_print("let x: Length = 0", "let x: Length = 0");
+    expect_pretty_print("let x = 0", "let x: forall A: Dim. A = 0"); // TODO: This is not ideal. 'forall' is not valid Numbat syntax.
+
+    // Derived unit definitions
+    expect_pretty_print(
+        "unit my_length_base_unit: Length",
+        "unit my_length_base_unit: Length",
+    );
+    expect_pretty_print(
+        "unit my_custom_base_unit",
+        "unit my_custom_base_unit: MyCustomBaseUnit",
+    );
+    expect_pretty_print(
+        "unit my_speed_unit = 10 m/s",
+        "unit my_speed_unit: Velocity = 10 metre / second",
+    );
+
+    // Function definitions
+    expect_pretty_print(
+        "fn f(v)=(v+1 knot)/1s",
+        "fn f(v: Velocity) -> Acceleration = (v + 1 knot) / 1 second",
+    );
+
+    expect_pretty_print("fn f(x) = x", "fn f<A>(x: A) -> A = x");
+
+    expect_pretty_print("fn f(x, y) = y", "fn f<A, B>(x: A, y: B) -> B = y");
+    expect_pretty_print("fn f(x, y) = x", "fn f<A, B>(x: B, y: A) -> B = x"); // TODO: This is correct, but it would be nice to associate 'x' to 'A', not 'B'.
+
+    expect_pretty_print("fn f(x) = 2 x", "fn f<A: Dim>(x: A) -> A = 2 x");
+
+    // Partially annotated functions
+    expect_pretty_print(
+        "fn f() -> Length * Frequency = c",
+        "fn f() -> Length × Frequency = c",
+    );
+    expect_pretty_print(
+        "fn f(v: Length * Frequency) = 1",
+        "fn f(v: Length × Frequency) -> Scalar = 1",
+    );
+
+    expect_pretty_print("fn f(x: Length) = 2 x", "fn f(x: Length) -> Length = 2 x");
+    expect_pretty_print("fn f(x) -> Length = 2 x", "fn f(x: Length) -> Length = 2 x");
+
+    // TODO:
+    // expect_pretty_print("fn f<Z>(z: Z) = z", "fn f<Z>(z: Z) -> Z = z");
 }
